import importlib
import xml.etree.ElementTree as ET
import logging
from typing import Callable, Dict

from medical.models import Item, Service

import core
from core.models import Officer
from core.utils import filter_validity
from django.db import connection, transaction
from gettext import gettext as _

from core.signals import register_service_signal
from .apps import ClaimConfig
from django.conf import settings

from claim.models import Claim, ClaimItem, ClaimService, ClaimDetail, ClaimDedRem, FeedbackPrompt
from product.models import ProductItemOrService

from claim.utils import process_items_relations, process_services_relations
from .validations import validate_claim, validate_assign_prod_to_claimitems_and_services, process_dedrem, \
    approved_amount, get_claim_category
from django.db.models import Subquery, F, OuterRef, Sum, FloatField
from django.db.models.functions import Coalesce
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied, ValidationError, ObjectDoesNotExist
logger = logging.getLogger(__name__)


@core.comparable
class ClaimElementSubmit(object):
    def __init__(self, type, code, quantity, price=None):
        self.type = type
        self.code = code
        self.price = price
        self.quantity = quantity

    def add_to_xmlelt(self, xmlelt):
        item = ET.SubElement(xmlelt, self.type)
        ET.SubElement(item, "%sCode" % self.type).text = "%s" % self.code
        if self.price:
            ET.SubElement(item, "%sPrice" % self.type).text = "%s" % self.price
        ET.SubElement(item, "%sQuantity" %
                      self.type).text = "%s" % self.quantity

    def to_claim_provision(self):
        raise NotImplementedError()


@core.comparable
class ClaimItemSubmit(ClaimElementSubmit):
    def __init__(self, code, quantity, price=None):
        super().__init__(type='Item',
                         code=code,
                         price=price,
                         quantity=quantity)

    def to_claim_provision(self):
        item = Item.objects.filter(validity_to__isnull=True, code=self.code).get()
        return ClaimItem(qty_provided=self.quantity, price_asked=self.price, item=item)


@core.comparable
class ClaimServiceSubmit(ClaimElementSubmit):
    def __init__(self, code, quantity, price=None):
        super().__init__(type='Service',
                         code=code,
                         price=price,
                         quantity=quantity)

    def to_claim_provision(self):
        service = Service.objects.filter(validity_to__isnull=True, code=self.code).get()
        return ClaimService(qty_provided=self.quantity, price_asked=self.price, service=service)


@core.comparable
class ClaimSubmit(object):
    def __init__(self, date, code, icd_code, total, start_date,
                 insuree_chf_id, health_facility_code,
                 claim_admin_code,
                 item_submits=None, service_submits=None,
                 end_date=None,
                 icd_code_1=None, icd_code_2=None, icd_code_3=None, icd_code_4=None,
                 visit_type=None, guarantee_no=None,
                 comment=None,
                 ):
        self.date = date
        self.code = code
        self.icd_code = icd_code
        self.total = total
        self.start_date = start_date
        self.insuree_chf_id = insuree_chf_id
        self.health_facility_code = health_facility_code
        self.end_date = end_date
        self.icd_code_1 = icd_code_1
        self.icd_code_2 = icd_code_2
        self.icd_code_3 = icd_code_3
        self.icd_code_4 = icd_code_4
        self.claim_admin_code = claim_admin_code
        self.visit_type = visit_type
        self.guarantee_no = guarantee_no
        self.comment = comment
        self.items = item_submits
        self.services = service_submits

    def _details_to_xmlelt(self, xmlelt):
        ET.SubElement(xmlelt, 'ClaimDate').text = self.date.to_ad_date().strftime(
            "%d/%m/%Y")
        ET.SubElement(
            xmlelt, 'HFCode').text = "%s" % self.health_facility_code
        if self.claim_admin_code:
            ET.SubElement(
                xmlelt, 'ClaimAdmin').text = "%s" % self.claim_admin_code
        ET.SubElement(xmlelt, 'ClaimCode').text = "%s" % self.code
        ET.SubElement(xmlelt, 'CHFID').text = "%s" % self.insuree_chf_id
        ET.SubElement(
            xmlelt, 'StartDate').text = self.start_date.to_ad_date().strftime("%d/%m/%Y")
        if self.end_date:
            ET.SubElement(xmlelt, 'EndDate').text = self.end_date.to_ad_date().strftime(
                "%d/%m/%Y")
        ET.SubElement(xmlelt, 'ICDCode').text = "%s" % self.icd_code
        if self.comment:
            ET.SubElement(xmlelt, 'Comment').text = "%s" % self.comment
        ET.SubElement(xmlelt, 'Total').text = "%s" % self.total
        if self.icd_code_1:
            ET.SubElement(xmlelt, 'ICDCode1').text = "%s" % self.icd_code_1
        if self.icd_code_2:
            ET.SubElement(xmlelt, 'ICDCode2').text = "%s" % self.icd_code_2
        if self.icd_code_3:
            ET.SubElement(xmlelt, 'ICDCode3').text = "%s" % self.icd_code_3
        if self.icd_code_4:
            ET.SubElement(xmlelt, 'ICDCode4').text = "%s" % self.icd_code_4
        if self.visit_type:
            ET.SubElement(xmlelt, 'VisitType').text = "%s" % self.visit_type
        if self.guarantee_no:
            ET.SubElement(
                xmlelt, 'GuaranteeNo').text = "%s" % self.guarantee_no

    def add_elt_list_to_xmlelt(self, xmlelt, elts_name, elts):
        if elts and len(elts) > 0:
            elts_xml = ET.SubElement(xmlelt, elts_name)
            for item in elts:
                item.add_to_xmlelt(elts_xml)

    def add_to_xmlelt(self, xmlelt):
        details = ET.SubElement(xmlelt, 'Details')
        self._details_to_xmlelt(details)
        self.add_elt_list_to_xmlelt(xmlelt, 'Items', self.items)
        self.add_elt_list_to_xmlelt(xmlelt, 'Services', self.services)

    def to_xml(self):
        claim_xml = ET.Element('Claim')
        self.add_to_xmlelt(claim_xml)
        return ET.tostring(claim_xml, encoding='utf-8', method='xml').decode()


@core.comparable
class ClaimSubmitError(Exception):
    ERROR_CODES = {
        -1: "Fatal Error",
        1: "Invalid HF Code",
        2: "Duplicate Claim Code",
        3: "Invalid Insuree CHFID",
        4: "End date is smaller than start date",
        5: "Invalid ICDCode",
        6: "Claimed amount is 0",
        7: "Invalid ItemCode",
        8: "Invalid ServiceCode",
        9: "Invalid Claim Admin",
    }

    def __init__(self, code, msg=None):
        self.code = code
        self.msg = ClaimSubmitError.ERROR_CODES.get(
            self.code, msg or "Unknown exception")

    def __str__(self):
        return "ClaimSubmitError %s: %s" % (self.code, self.msg)


class ClaimSubmitService(object):

    def __init__(self, user):
        self.user = user

    def hf_scope_check(self, claim_submit: ClaimSubmit):
        self._validate_user_hf(claim_submit.health_facility_code)

    def submit(self, claim_submit):
        with connection.cursor() as cur:
            sql = """\
                DECLARE @ret int;
                EXEC @ret = [dbo].[uspUpdateClaimFromPhone] @XML = %s;
                SELECT @ret;
            """

            cur.execute(sql, (claim_submit.to_xml(),))
            for i in range(int(ClaimConfig.claim_uspUpdateClaimFromPhone_intermediate_sets)):
                cur.nextset()
            if cur.description is None:  # 0 is considered as 'no result' by pyodbc
                return
            res = cur.fetchone()[0]  # FETCH 'SELECT @ret' returned value
            raise ClaimSubmitError(res)

    @register_service_signal('claim.enter_and_submit_claim')
    @transaction.atomic
    def enter_and_submit(self, claim: dict, rule_engine_validation: bool = True) -> Claim:
        create_claim_service = ClaimCreateService(self.user)
        entered_claim = create_claim_service.enter_claim(claim)
        submitted_claim = self.submit_claim(entered_claim, rule_engine_validation)
        return submitted_claim

    @register_service_signal('claim.submit_claim')
    def submit_claim(self, claim: Claim, rule_engine_validation=True):
        """
        Submission based on the GQL SubmitClaimMutation.async_mutate
        """
        self._validate_submit_permissions()
        self._validate_user_hf(claim.health_facility.code)
        claim.save_history()

        if rule_engine_validation:
            validation_errors = self._validate_claim(claim)
            if validation_errors:
                return self.__submit_to_rejected(claim)

        return self.__submit_to_checked(claim)

    def _validate_submit_permissions(self):
        if type(self.user) is AnonymousUser or not self.user.id:
            raise ValidationError(
                _("mutation.authentication_required"))
        if not self.user.has_perms(ClaimConfig.gql_mutation_submit_claims_perms):
            raise PermissionDenied(_("unauthorized"))

    def _validate_user_hf(self, hf_code):
        from location.models import LocationManager, HealthFacility
        hf = LocationManager().build_user_location_filter_query(self.user._u, queryset = HealthFacility.filter_queryset().filter(code=hf_code))
        if not hf and settings.ROW_SECURITY:
            raise ClaimSubmitError("Invalid health facility code or health facility not allowed for user")

    def _validate_claim(self, claim):
        errors = validate_claim(claim, True)
        if not errors:
            errors = validate_assign_prod_to_claimitems_and_services(claim)
            errors += process_dedrem(claim, self.user.id_for_audit, False)
        return errors or []

    def __submit_to_rejected(self, claim: Claim):
        claim.status = Claim.STATUS_REJECTED
        claim.save()
        return claim

    def __submit_to_checked(self, claim: Claim):
        claim.approved = approved_amount(claim)
        claim.status = Claim.STATUS_CHECKED
        from core.utils import TimeUtils
        claim.submit_stamp = TimeUtils.now()
        claim.category = get_claim_category(claim)
        claim.save()
        return claim


def formatClaimService(s):
    return {
        "service": str(s.service),
        "quantity": s.qty_provided,
        "price": s.price_asked,
        "explanation": s.explanation
    }


def formatClaimItem(i):
    return {
        "item": str(i.item),
        "quantity": i.qty_provided,
        "price": i.price_asked,
        "explanation": i.explanation
    }


class ClaimReportService(object):
    def __init__(self, user):
        self.user = user

    def fetch(self, uuid):
        from .models import Claim
        queryset = Claim.objects.filter(*core.filter_validity())
        if settings.ROW_SECURITY:
            from location.models import LocationManager
            queryset = LocationManager().build_user_location_filter_query( self.user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])
        claim = queryset\
            .select_related('health_facility') \
            .select_related('insuree') \
            .filter(uuid=uuid)\
            .first()
        if not claim:
            raise PermissionDenied(_("unauthorized"))
        return {
            "code": claim.code,
            "visitDateFrom": claim.date_from.isoformat() if claim.date_from else None,
            "visitDateTo":  claim.date_to.isoformat() if claim.date_to else None,
            "claimDate": claim.date_claimed.isoformat() if claim.date_claimed else None,
            "healthFacility": str(claim.health_facility),
            "insuree": str(claim.insuree),
            "claimAdmin": str(claim.admin) if claim.admin else None,
            "icd": str(claim.icd),
            "icd1": str(claim.icd1) if claim.icd_1 else None,
            "icd2": str(claim.icd1) if claim.icd_2 else None,
            "icd3": str(claim.icd1) if claim.icd_3 else None,
            "icd4": str(claim.icd1) if claim.icd_4 else None,
            "guarantee": claim.guarantee_id,
            "visitType": claim.visit_type,
            "claimed": claim.claimed,
            "services": [formatClaimService(s) for s in claim.services.all()],
            "items": [formatClaimItem(i) for i in claim.items.all()],
        }


class ClaimCreateService:
    def __init__(self, user):
        self.user = user

    def _validate_user_hf(self, hf_id):
        from location.models import LocationManager, HealthFacility
        hf = LocationManager().build_user_location_filter_query(self.user._u, queryset = HealthFacility.filter_queryset().filter(id=hf_id))
        if not hf and settings.ROW_SECURITY:
            raise ValidationError("Invalid health facility code or health facility not allowed for user")

    @register_service_signal('claim.enter_claim')
    def enter_claim(self, claim: dict):
        """
        Implementation based on the GQL CreateClaimMutation.async_mutate
        """
        self._validate_permissions()
        self._validate_claim_fields(claim)
        self._validate_user_hf(claim.get('health_facility_id', None))
        self._ensure_entered_claim_fields(claim)
        claim = self._create_claim_from_dict(claim)
        return claim

    def _validate_permissions(self):
        if type(self.user) is AnonymousUser or not self.user.id:
            raise ValidationError(
                _("mutation.authentication_required"))
        if not self.user.has_perms(ClaimConfig.gql_mutation_create_claims_perms):
            raise PermissionDenied(_("unauthorized"))

    def _validate_claim_fields(self, claim):
        if not claim.get('code'):
            raise ValidationError("Provided claim without code.")

        if Claim.objects.filter(code=claim['code'], validity_to__isnull=True).exists():
            raise ValidationError(F"Claim with code '{claim['code']}' already exists.")

    def _ensure_entered_claim_fields(self, claim_submit_data):
        claim_submit_data['audit_user_id'] = self.user.id_for_audit
        claim_submit_data['status'] = Claim.STATUS_ENTERED
        from core.utils import TimeUtils
        claim_submit_data['validity_from'] = TimeUtils.now()

    def _create_claim_from_dict(self, claim_submit_data):
        items = claim_submit_data.pop('items', [])
        services = claim_submit_data.pop('services', [])
        claim_submit_data.pop('service_item_set', [])
        claim_submit_data.pop('service_service_set', [])
        claim = Claim.objects.create(**claim_submit_data)
        self.__process_items(claim, items, services)
        claim.save()
        return claim

    def __process_items(self, claim, items, services):
        claimed = 0
        claimed += process_items_relations(self.user, claim, items)
        claimed += process_services_relations(self.user, claim, services)
        claim.claimed = claimed


def update_sum_claims(claim):
    claimed = 0
    service_asked = Subquery(
        ClaimItem.objects.filter(claim=OuterRef('pk')).filter(legacy_id__isnull=True).values('claim_id').annotate(
            item_sum=Sum(F('price_asked')*F('qty_provided'))).values('item_sum').order_by()[:1],
        output_field=FloatField()
    )
    item_asked = Subquery(
        ClaimService.objects.filter(claim=OuterRef('pk')).filter(legacy_id__isnull=True).values('claim_id').annotate(
            service_sum=Sum(F('price_asked')*F('qty_provided'))).values('service_sum').order_by()[:1],
        output_field=FloatField()
    )
    Claim.objects.filter(id=claim.id).update(
        claimed=Coalesce(item_asked, 0) + Coalesce(service_asked, 0)
    )


def check_unique_claim_code(code):
    if Claim.objects.filter(code=code, validity_to__isnull=True).exists():
        return [{"message": "Claim code %s already exists" % code}]
    return []


def reset_claim_before_update(claim):
    claim.date_to = None
    claim.icd_1 = None
    claim.icd_2 = None
    claim.icd_3 = None
    claim.icd_4 = None
    claim.guarantee_id = None
    claim.explanation = None
    claim.adjustment = None
    claim.json_ext = None


def __autogenerate_claim_code():
    module_name, function_name = '[undefined]', '[undefined]'
    try:
        claim_code_function = _get_autogenerating_func()
        return claim_code_function(ClaimConfig.autogenerated_claim_code_config)
    except ImportError as e:
        logger.error(f"Error: Could not import module '{module_name}' for claim code autogeneration")
        raise e
    except AttributeError as e:
        logger.error(f"Error: Could not find function '{function_name}' in module '{module_name}' for claim code autogeneration")
        raise e


def _get_autogenerating_func() -> Callable[[Dict], Callable]:
    module_name, function_name = ClaimConfig.autogenerate_func.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, function_name)


def claim_create(data, user, autogenerate_code = False):
    restore = data.pop('restore', None)
    autogenerate_code = data.pop('autogenerate', None)
    if restore:
        data["restore"] = Claim.objects.filter(uuid=restore).first()
    
    if autogenerate_code:
        data['code'] = __autogenerate_claim_code()
    data['audit_user_id'] = user.id_for_audit
    claim = Claim()
    set_reduced_attr(claim, data, ['items', 'services'])
    claim.save()
    claim_create_items_and_services(claim, data, user)
    return claim


def claim_update(claim, data, user):
    claim.save_history()
    # reset the non required fields
    # (each update is 'complete', necessary to be able to set 'null')
    reset_claim_before_update(claim)
    set_reduced_attr(claim, data, ['items', 'services'])
    from core.utils import TimeUtils
    claim.items.update(validity_to=TimeUtils.now())
    claim.services.update(validity_to=TimeUtils.now())
    claim_create_items_and_services(claim, data, user)
    return claim


def set_reduced_attr(obj, data, exclusions):
    for key in data:
        if key not in exclusions:
            setattr(obj,key, data[key]) 


def claim_create_items_and_services(claim, data, user):
    items = data.pop('items') if 'items' in data else []
    services = data.pop('services') if 'services' in data else []
    claimed = 0
    claimed += process_items_relations(user, claim, items)
    claimed += process_services_relations(user, claim, services)
    claim.claimed = claimed
    claim.save()


def update_or_create_claim(data, user):
         
    validate_claim_data(data, user)
    claim_uuid = data.pop("uuid", None)
    # update_or_create(uuid=claim_uuid, ...)
    # doesn't work because of explicit attempt to set null to uuid!
   
    if claim_uuid:
        claim = Claim.objects.get(uuid=claim_uuid)
        claim = claim_update(claim, data, user)
    else:
        claim = claim_create(data, user)
    return claim


def validate_claim_data(data, user):
    services = data.get('services') if 'services' in data else []
    incoming_code = data.get('code')
    claim_uuid = data.get("uuid", None)
    restore = data.get('restore', None)
    current_claim = Claim.objects.filter(uuid=claim_uuid).first()
    current_code = current_claim.code if current_claim else None

    if restore:
        restored_qs = Claim.objects.filter(uuid=restore)
        restored_from_claim = restored_qs.first()
        restored_count = Claim.objects.filter(restore=restored_from_claim).count()
        if not restored_qs.exists():
            raise ValidationError(_("mutation.restored_from_does_not_exist"))
        if not restored_from_claim.status == Claim.STATUS_REJECTED:
            raise ValidationError(_("mutation.cannot_restore_not_rejected_claim"))
        if not user.has_perms(ClaimConfig.gql_mutation_restore_claims_perms):
            raise ValidationError(_("mutation.no_restore_rights"))
        if ClaimConfig.claim_max_restore and restored_count >= ClaimConfig.claim_max_restore:
            raise ValidationError(_("mutation.max_restored_claim") % {
                "max_restore": ClaimConfig.claim_max_restore
            })
           
    elif current_claim is not None and current_claim.status not in (Claim.STATUS_CHECKED, Claim.STATUS_ENTERED):
        raise ValidationError(_("mutation.claim_not_editable")) 

    if not validate_number_of_additional_diagnoses(data):
        raise ValidationError(_("mutation.claim_too_many_additional_diagnoses"))

    if ClaimConfig.claim_validation_multiple_services_explanation_required:
        for service in services:
            if service["qty_provided"] > 1 and not service.get("explanation"):
                raise ValidationError(_("mutation.service_explanation_required"))

    if len(incoming_code) > ClaimConfig.max_claim_length:
        raise ValidationError(_("mutation.code_name_too_long"))

    if not restore and current_code != incoming_code and check_unique_claim_code(incoming_code):
        raise ValidationError(_("mutation.code_name_duplicated"))


def validate_number_of_additional_diagnoses(incoming_data):
    additional_diagnoses_count = 0
    for key in incoming_data.keys():
        if key.startswith("icd_") and key.endswith("_id") and key != "icd_id":
            additional_diagnoses_count += 1

    return additional_diagnoses_count <= ClaimConfig.additional_diagnosis_number_allowed


def submit_claim(claim, user):
    c_errors = []
    claim.save_history()
    logger.debug("SubmitClaimsMutation: validating claim %s", claim.uuid)
    c_errors += validate_claim(claim, True)
    logger.debug("SubmitClaimsMutation: claim %s validated, nb of errors: %s", claim.uuid, len(c_errors))
    if len(c_errors) == 0:
        c_errors = validate_assign_prod_to_claimitems_and_services(claim)
        logger.debug("SubmitClaimsMutation: claim %s assigned, nb of errors: %s", claim.uuid, len(c_errors))
        c_errors += process_dedrem(claim, user.id_for_audit, False)
        logger.debug("SubmitClaimsMutation: claim %s processed for dedrem, nb of errors: %s", claim.uuid,
                        len(c_errors))
    c_errors += set_claim_submitted(claim, c_errors, user)
    logger.debug("SubmitClaimsMutation: claim %s set submitted", claim.uuid)
    return c_errors


def set_claim_submitted(claim, errors, user):
    try:
        claim.audit_user_id_submit = user.id_for_audit
        if errors:
            claim.status = Claim.STATUS_REJECTED
        else:
            claim.approved = approved_amount(claim)
            claim.status = Claim.STATUS_CHECKED
            from core.utils import TimeUtils
            claim.submit_stamp = TimeUtils.now()
            claim.category = get_claim_category(claim)
        claim.save()
        return []
    except Exception as exc:
        return {
            'title': claim.code,
            'list': [{
                'message': _("claim.mutation.failed_to_change_status_of_claim") % {'code': claim.code},
                'detail': claim.uuid}]
        }
        

def validate_and_process_dedrem_claim(claim, user, is_process):
    errors = validate_claim(claim, False)
    logger.debug("ProcessClaimsMutation: claim %s validated, nb of errors: %s", claim.uuid, len(errors))
    if len(errors) == 0:
        errors = validate_assign_prod_to_claimitems_and_services(claim)
        logger.debug("ProcessClaimsMutation: claim %s assigned, nb of errors: %s", claim.uuid, len(errors))
        errors += process_dedrem(claim, user.id_for_audit, is_process)
        logger.debug("ProcessClaimsMutation: claim %s processed for dedrem, nb of errors: %s", claim.uuid,
                     len(errors))
    else:
        # OMT-208 the claim is invalid. If there is a dedrem, we need to clear it (caused by a review)
        deleted_dedrems = ClaimDedRem.objects.filter(claim=claim).delete()
        if deleted_dedrems:
            logger.debug(f"Claim {claim.uuid} is invalid, we deleted its dedrem ({deleted_dedrems})")
    if is_process:
        errors += set_claim_processed_or_valuated(claim, errors, user)
    return errors


def set_claim_processed_or_valuated(claim, errors, user):
    try:
        if errors:
            claim.status = Claim.STATUS_REJECTED
        else:
            claim.status = Claim.STATUS_PROCESSED if with_relative_prices(claim) else Claim.STATUS_VALUATED
            claim.audit_user_id_process = user.id_for_audit
            from core.utils import TimeUtils
            claim.process_stamp = TimeUtils.now()
        claim.save()
        return []
    except Exception as ex:
        error = {
            'title': claim.code,
            'list': [{'message': _("claim.mutation.failed_to_change_status_of_claim") % {'code': claim.code},
                      'detail': claim.uuid}]
        }
        if hasattr(ex, 'args') and len(ex.args)>0:
            for arg in ex.args:
            
                error['list'].append(arg)
        return [error]


def details_with_relative_prices(details):
    return details.filter(status=ClaimDetail.STATUS_PASSED) \
        .filter(price_origin=ProductItemOrService.ORIGIN_RELATIVE) \
        .exists()


def with_relative_prices(claim):
    return details_with_relative_prices(claim.items) or details_with_relative_prices(claim.services)


def set_claims_status(uuids, field, status, audit_data=None, user=None):
    errors = []
    claims = Claim.objects \
            .filter(uuid__in=uuids,
                    *filter_validity()) 
    remaining_uuid = list(map(str.upper,uuids))
    for claim in claims:
        remaining_uuid.remove(claim.uuid.upper())        
        try:
            claim.save_history()
            setattr(claim, field, status)
            # creating/cancelling feedback prompts
            if field == 'feedback_status':
                if status == Claim.FEEDBACK_SELECTED:
                    create_feedback_prompt(claim, user)
                elif status in [Claim.FEEDBACK_NOT_SELECTED, Claim.FEEDBACK_BYPASSED]:
                    set_feedback_prompt_validity_to_to_current_date(claim.uuid)
            if audit_data:
                for k, v in audit_data.items():
                    setattr(claim, k, v)
            claim.save()
        except Exception as exc:
            errors += [
                {'message': _("claim.mutation.failed_to_change_status_of_claim") %
                            {'code': claim.code} }
            ]
            if hasattr(exc, 'messages') and len(exc.messages):
                for m in exc.messages:
                    errors.append({'message': m })
            elif hasattr(exc, 'args') and len(exc.args):
                for m in exc.args:
                    errors.append({'message': m })
        if len(remaining_uuid):
            errors.append(_(
                "claim.validation.id_does_not_exist") % {'id': ','.join(remaining_uuid)})                  
                        
    return errors


def create_feedback_prompt(current_claim, user):
    
    feedback_prompt = {}
    from core.utils import TimeUtils
    feedback_prompt['feedback_prompt_date'] = TimeUtils.date()
    feedback_prompt['validity_from'] = TimeUtils.now()
    feedback_prompt['claim'] = current_claim
    villages = []
    if current_claim.insuree.current_village:
        villages.append(current_claim.insuree.current_village)
    if current_claim.insuree.family.location:
        villages.append(current_claim.insuree.family.location)
    officer = Officer.objects.filter(
        *filter_validity(),
        officer_villages__location__in=villages,
        phone__isnull = False
    ).exclude(phone__exact= '').first()
    if not officer:
        bad_officer = Officer.objects.filter(
        *filter_validity(),
        officer_villages__location__in=villages,
        ).first()
        if bad_officer:
            msg = [' officer '+bad_officer.code+ 'has not phone setup']
        else:
            msg = []
        raise RuntimeError(f"No officer with a phone number found for the insuree village code, \
            {', '.join([str(v.code) for v in villages ])}", *msg)
            
    feedback_prompt['officer_id'] = officer.id
    feedback_prompt['phone_number'] = officer.phone
    feedback_prompt['audit_user_id'] = user.id_for_audit
    FeedbackPrompt.objects.create(
        **feedback_prompt
    )


def set_feedback_prompt_validity_to_to_current_date(claim_uuid):
    try:
        claim = Claim.objects.get(uuid=claim_uuid).id
        feedback_prompt_id = FeedbackPrompt.objects.get(claim=claim, validity_to=None).id
        from core.utils import TimeUtils
        current_feedback_prompt = FeedbackPrompt.objects.get(id=feedback_prompt_id)
        current_feedback_prompt.validity_to = TimeUtils.now()
        current_feedback_prompt.save()
    except ObjectDoesNotExist:
        return "No such feedback prompt exist."


def update_claims_dedrems(uuids, user):
    # We could do it in one query with filter(claim__uuid__in=uuids) but we'd loose the logging
    errors = []
    claims = Claim.objects.filter(uuid__in=uuids)
    remaining_uuid = list(map(str.upper,uuids))
    for claim in claims:
        remaining_uuid.remove(claim.uuid.upper())       
        logger.debug(f"delivering review on {claim.uuid}, reprocessing dedrem ({user})")
        errors += validate_and_process_dedrem_claim(claim, user, False)
    if len(remaining_uuid):
        errors.append(_(
            "claim.validation.id_does_not_exist") % {'id': ','.join(remaining_uuid)})
    return errors



import random
import string

from claim.models import ClaimAdmin
from claim.test_helpers import create_test_claim, create_test_claimservice, create_test_claimitem
from django.core.management.base import BaseCommand
from insuree.models import Insuree
from medical.models import Service, Item


class Command(BaseCommand):
    help = "This command will generate test Claims with some optional parameters. It is intended to simulate larger" \
           "databases for performance testing"
    insurees = None
    services = None
    items = None
    claim_admins = None
    hfs = None

    def add_arguments(self, parser):
        parser.add_argument("nb_claims", nargs=1, type=int)
        parser.add_argument("nb_services", nargs=1, type=int, help="number of services per claim, with 10% randomness")
        parser.add_argument("nb_items", nargs=1, type=int, help="number of items per claim, with 10% randomness")
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            help='Be verbose about what it is doing',
        )

    def handle(self, *args, **options):
        nb_claims = options["nb_claims"][0]
        nb_services = options["nb_services"][0]
        nb_items = options["nb_items"][0]
        verbose = options["verbose"]
        for claim_num in range(1, nb_claims + 1):
            insuree = self.get_random_insuree()
            claim_admin = self.get_random_claim_admin()
            hf = self.get_random_hf()
            claim = create_test_claim({"insuree_id": insuree,
                                       "code": ''.join(random.choices(string.ascii_uppercase + string.digits, k=8)),
                                       "admin_id": claim_admin,
                                       "health_facility_id": hf})
            if verbose:
                logger.debug(f"{claim_num} created claim {claim} for insuree {insuree} with code {claim.code}")
            for svc_num in range(1, nb_services + 1):
                service = self.get_random_service()
                claim_service = create_test_claimservice(claim, custom_props={
                    "service_id": service,
                    "qty_provided": random.randint(1, 10),
                    "price_asked": random.randint(1, 1000),
                })
                if verbose:
                    logger.debug(f"{claim_num} {svc_num} Created claim service {claim_service} for service {service}")
            for item_num in range(1, nb_items + 1):
                item = self.get_random_item()
                claim_item = create_test_claimitem(claim, "D", custom_props=
                {
                    "item_id": item,
                    "qty_provided": random.randint(1, 10),
                    "price_asked": random.randint(1, 1000),
                })
                if verbose:
                    logger.debug(f"{claim_num} {item_num} Created claim item {claim_item} for item {item}")

    def get_random_insuree(self):
        if not self.insurees:
            self.insurees = Insuree.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.insurees)

    def get_random_service(self):
        if not self.services:
            self.services = Service.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.services)

    def get_random_item(self):
        if not self.items:
            self.items = Item.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.items)

    def get_random_claim_admin(self):
        if not self.claim_admins:
            self.claim_admins = ClaimAdmin.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.claim_admins)

    def get_random_hf(self):
        if not self.hfs:
            from location.models import HealthFacility
            self.hfs = HealthFacility.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.hfs)



import os
import importlib

dir_path = os.path.dirname(__file__)
for file in os.listdir(dir_path):
    if file.endswith('.py') and file != '__init__.py':
        module_name = file[:-3]  # we only care about name of the file, not extension
        module = importlib.import_module(f'.{module_name}', package=__name__)
# dynamic import from attachment_strategies directory, this will be marked as unresolved reference by most linters
attachment_strategies_dict = {"claimdoc": handle_claimdoc_attachment}



from claim.models import Claim, ClaimService, ClaimItem, ClaimDedRem, ClaimAdmin
from claim.validations import get_claim_category, approved_amount
from claim.services import claim_create, update_sum_claims
from medical.test_helpers import get_item_of_type, get_service_of_category
from uuid import uuid4

class DummyUser:
    def __init__(self):
      self.id_for_audit = 1  

def create_test_claim(custom_props={}, user = DummyUser() ):
    from core import datetime
    if 'insuree' not in custom_props and 'insuree_id' not in custom_props:
        custom_props["insuree_id"]= 2

    return claim_create(
        {
            "health_facility_id": 18,
            "icd_id": 116,
            "date_from": datetime.datetime(2019, 6, 1),
            "date_claimed": datetime.datetime(2019, 6, 1),
            "date_to": datetime.datetime(2019, 6, 1),
            "status": 2,
            "validity_from": datetime.datetime(2019, 6, 1),
            "code":  str(uuid4()),
            **custom_props
        }, user
    )


def create_test_claimitem(claim, item_type, valid=True, custom_props={}):
    item =  ClaimItem.objects.create(
        **{
            "claim": claim,
            "qty_provided": 7,
            "price_asked": 11,
            "item_id": get_item_of_type(item_type).id if item_type else 23,  # Atropine
            "status": 1,
            "availability": True,
            "validity_from": "2019-06-01",
            "validity_to": None if valid else "2019-06-01",
            "audit_user_id": -1,
            **custom_props
           }
    )
    update_sum_claims(claim)
    return item



def create_test_claimservice(claim, category=None, valid=True, custom_props={}):
    service =  ClaimService.objects.create(
        **{
            "claim": claim,
            "qty_provided": 7,
            "price_asked": 11,
            "service_id": get_service_of_category(category).id if category else 23,  # Skin graft, no cat
            "status": 1,
            "validity_from": "2019-06-01",
            "validity_to": None if valid else "2019-06-01",
            "audit_user_id": -1,
            **custom_props
        }
    )    
    update_sum_claims(claim)
    return service



def mark_test_claim_as_processed(claim, status=Claim.STATUS_CHECKED, audit_user_id=-1):
    claim.approved = approved_amount(claim)
    claim.status = status
    claim.audit_user_id_submit = audit_user_id
    from core.utils import TimeUtils
    claim.submit_stamp = TimeUtils.now()
    claim.category = get_claim_category(claim)
    claim.save()


def delete_claim_with_itemsvc_dedrem_and_history(claim):
    # first delete old versions of the claim
    ClaimDedRem.objects.filter(claim=claim).delete()
    old_claims = Claim.objects.filter(legacy_id=claim.id)
    ClaimItem.objects.filter(claim__in=old_claims).delete()
    ClaimService.objects.filter(claim__in=old_claims).delete()
    old_claims.delete()
    claim.items.all().delete()
    claim.services.all().delete()
    claim.delete()


def create_test_claim_admin(custom_props={}):
    from core import datetime
    code = custom_props.pop('code','TST-CA')
    uuid = custom_props.pop('uuid',None)
    ca = None
    qs_ca = ClaimAdmin.objects
    data = {
        "code": code,
        "uuid": uuid,
        "last_name": "LastAdmin",
        "other_names": "JoeAdmin",
        "email_id": "joeadmin@lastadmin.com",
        "phone": "+12027621401",
        "health_facility_id": 1,
        "has_login": False,
        "audit_user_id": 1,
        "validity_from": datetime.datetime(2019, 6, 1),
        **custom_props
    }
    if code:
        qs_ca = qs_ca.filter(code=code)
    if uuid:
        qs_ca = qs_ca.filter(uuid=uuid)
        
    if code or uuid:
        ca = qs_ca.first()
    if ca:
        data['uuid']=ca.uuid
        ca.update(data)
        return ca
    else:
        data['uuid']=uuid4()
        return ClaimAdmin.objects.create( **data)



# Generated by Django 3.2.16 on 2023-08-22 07:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0021_auto_20230818_0914'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feedbackprompt',
            name='phone_number',
            field=models.CharField(db_column='PhoneNumber', max_length=50, null=True, blank=True),
        ),
    ]



# Generated by Django 2.1.14 on 2020-03-18 14:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0006_claimattachment_url'),
    ]

    operations = [
        migrations.AlterField(
            model_name='claimattachment',
            name='legacy_id',
            field=models.UUIDField(blank=True, db_column='LegacyID', null=True),
        ),
    ]



# Generated by Django 3.2.16 on 2023-06-15 10:55

from django.db import migrations, models
import django.db.models.deletion
from claim.models import Claim


class Migration(migrations.Migration):

    dependencies = [
        ('location', '0015_set_managed_to_true'),
        ('claim', '0018_alter_jsonext_column'),
    ]

    operations = []
    try:
        Claim.objects.filter(pk__lt=10).aggregate(sum=models.Count('refer_from'))
    except:
        operations.append(migrations.AddField(
            model_name='claim',
            name='refer_from',
            field=models.ForeignKey(blank=True, db_column='ReferFrom', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='referFromHF', to='location.healthfacility'),
        ))
    try:
        Claim.objects.filter(pk__lt=10).aggregate(sum=models.Count('refer_to'))
    except:    
        operations.append(migrations.AddField(
            model_name='claim',
            name='refer_to',
            field=models.ForeignKey(blank=True, db_column='ReferTo', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='referToHF', to='location.healthfacility'),
        ))
    



import base64
import json
from dataclasses import dataclass
from core.models import User
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase
from core.utils import filter_validity
from core.test_helpers import create_test_interactive_user
from django.conf import settings
from graphene_django.utils.testing import GraphQLTestCase
from graphql_jwt.shortcuts import get_token
#credits https://docs.graphene-python.org/projects/django/en/latest/testing/
from claim import schema as claim_schema
from graphene.test import Client
from graphene import Schema

from claim.models import Claim, ClaimAdmin


from policy.models import Policy
from policy.test_helpers import create_test_policy2
from product.test_helpers import create_test_product, create_test_product_service
from core.test_helpers import create_test_officer
from insuree.test_helpers import create_test_insuree
from location.models import Location
from medical.test_helpers import create_test_service
from medical_pricelist.test_helpers import add_service_to_hf_pricelist

@dataclass
class DummyContext:
    """ Just because we need a context to generate. """
    user: User

class ClaimGraphQLTestCase(openIMISGraphQLTestCase):
    # This is required by some version of graphene but is never used. It should be set to the schema but the import
    # is shown as an error in the IDE, so leaving it as True.
    GRAPHQL_SCHEMA = True
    admin_user = None
    graph_client = None
    schema = None      
    officer= None
    insuree= None
    product= None
    service= None
    product_service= None
    claim_admin = None
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username="testLocationAdmin")
        cls.admin_token = get_token(cls.admin_user, DummyContext(user=cls.admin_user))
        cls.schema = Schema(
            query=claim_schema.Query,
            mutation=claim_schema.Mutation
        )
        cls.graph_client = Client(cls.schema)
        
        cls.officer = create_test_officer(custom_props={"code": "TSTSIMP1"})
        cls.insuree = create_test_insuree(custom_props={"chf_id": "paysimp"})
        cls.product = create_test_product("ELI1")
        (policy, insuree_policy) = create_test_policy2(cls.product, cls.insuree, custom_props={
            "value": 1000, "status": Policy.STATUS_ACTIVE})
        cls.service = create_test_service("A")
        cls.claim_admin = ClaimAdmin.objects.filter(*filter_validity()).first()
        cls.svc_pl_detail = add_service_to_hf_pricelist(cls.service, hf_id = cls.claim_admin.health_facility.id )
        cls.product_service = create_test_product_service(cls.product, cls.service, custom_props={"limit_no_adult": 20})
        
    def test_claims_query(self):
        
        response = self.query(
            '''
            query {
                claims
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            uuid,code,jsonExt,dateClaimed,dateProcessed,feedbackStatus,reviewStatus,claimed,approved,status,restoreId,healthFacility { id uuid name code },insuree{id, uuid, chfId, lastName, otherNames, dob},attachmentsCount
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)

        # Add some more asserts if you like
        ...

    def test_query_with_variables(self):
        response = self.query(
            '''
            query claims($status: Int!, $first:  Int! ) {
                claims(status: $status,orderBy: ["-dateClaimed"],first: $first)
                {
                    totalCount
                    pageInfo { hasNextPage, hasPreviousPage, startCursor, endCursor}
                    edges
                    {
                        node
                        {
                            uuid,code,jsonExt,dateClaimed,dateProcessed,feedbackStatus,reviewStatus,claimed,approved,status,restoreId,healthFacility { id uuid name code },insuree{id, uuid, chfId, lastName, otherNames, dob},attachmentsCount
                        }
                    }
                }
            }
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"},
            variables={'status': 2, 'first':10}
        )

        content = json.loads(response.content)

        # This validates the status code and if you get errors
        self.assertResponseNoErrors(response)
        
    def execute_mutation(self, mutation):
        mutation_result = self.graph_client.execute(mutation, context=DummyContext(user=self.admin_user))
        return mutation_result
        
    def test_mutation_create_claim(self):

        response = self.query(
            f'''
            mutation {{
                createClaim(
                    input: {{
                    clientMutationId: "3a90436a-d5ea-48e7-bde4-0bcff0240260"
                    clientMutationLabel: "Create Claim - m-c-claim" 
                    code: "m-c-claim"
                autogenerate: false
                insureeId: {self.insuree.id}
                adminId: {self.claim_admin.id}
                dateFrom: "2023-12-06"  
                icdId: 2 
                jsonExt: "{{}}"
                feedbackStatus: 1
                reviewStatus: 1
                dateClaimed: "2023-12-06"
                healthFacilityId: {self.claim_admin.health_facility.id}
                visitType: "O"
                services: [
                {{
                
                serviceId: {self.service.id}
                priceAsked: "10.00"
                qtyProvided: "1.00"
                status: 1,
                serviceItemSet: [],
                serviceServiceSet: []
            }}
                ]
                items: [
                ]
                    }}
                ) {{
                    clientMutationId
                    internalId
                }}
            }}
                ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        self.get_mutation_result('3a90436a-d5ea-48e7-bde4-0bcff0240260', self.admin_token )
        claim = Claim.objects.filter(code = 'm-c-claim').first()
        self.assertIsNotNone(claim)
        self.assertEqual(claim.status, Claim.STATUS_ENTERED)
        response = self.query(
            f'''
            mutation {{
                updateClaim(
                    input: {{
                    clientMutationId: "3a90436b-d5ea-48e7-bde4-0bcff0240260"
                    clientMutationLabel: "Update Claim - m-c-claim" 
                    code: "m-c-claim"
                autogenerate: false
                uuid: "{str(claim.uuid)}"
                insureeId: {self.insuree.id}
                adminId: {self.claim_admin.id}
                dateFrom: "2023-11-06"  
                icdId: 2 
                jsonExt: "{{}}"
                feedbackStatus: 1
                reviewStatus: 1
                dateClaimed: "2023-12-06"
                healthFacilityId: {self.claim_admin.health_facility.id}
                visitType: "O"
                services: [
                {{
                
                serviceId: {self.service.id}
                priceAsked: "10.00"
                qtyProvided: "1.00"
                status: 1,
                serviceItemSet: [],
                serviceServiceSet: []
            }}
                ]
                items: [
                ]
                    }}
                ) {{
                    clientMutationId
                    internalId
                }}
            }}
                ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        self.get_mutation_result('3a90436b-d5ea-48e7-bde4-0bcff0240260', self.admin_token )
                
        #submit claim 
        response = self.query(f'''
            mutation {{
            submitClaims(
                input: {{
                clientMutationId: "d02fff0a-dd95-4413-a2f4-4cf2189dc0d6"
                clientMutationLabel: "Submit claim erterwtw"
                
                uuids: ["{claim.uuid}"]
                }}
            ) {{
                clientMutationId
                internalId
            }}
            }}
            ''',
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        self.assertResponseNoErrors(response)
        self.get_mutation_result('d02fff0a-dd95-4413-a2f4-4cf2189dc0d6', self.admin_token )
        # select for feeback
        claim = Claim.objects.filter(code = 'm-c-claim').first()
        create_test_officer(villages=[claim.insuree.family.location])
        self.assertEqual(claim.status, Claim.STATUS_CHECKED)
        response = self.query(f'''
            mutation {{
            selectClaimsForFeedback(
                input: {{
                clientMutationId: "f0585e2b-d72d-4001-915a-1cf10e9f1722"
                clientMutationLabel: "Select claim sadddfas for feedback"
                
                uuids: ["{claim.uuid}"]
                }}
            ) {{
                clientMutationId
                internalId
            }}
            }}
        ''' ,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"})
        self.assertResponseNoErrors(response)
        self.get_mutation_result('f0585e2b-d72d-4001-915a-1cf10e9f1722', self.admin_token )
        ## check the mutation answer
        
        claim = Claim.objects.filter(code = 'm-c-claim').first()
        self.assertEqual(claim.feedback_status, Claim.FEEDBACK_SELECTED)


# Generated by Django 2.1.11 on 2019-10-29 18:49

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0008_officer_role_roleright_userrole'),
        ('claim', '0004_claimattachment'),
    ]

    operations = [
        migrations.CreateModel(
            name='ClaimMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('claim', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='claim.Claim')),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='claims', to='core.MutationLog')),
            ],
            options={
                'db_table': 'claim_ClaimMutation',
                'managed': True,
            },
        ),
    ]



# Generated by Django 3.2.18 on 2023-05-15 09:49

import core.fields
import datetime
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('insuree', '0013_auto_20211103_1023'),
        ('location', '0013_auto_20230317_1534'),
        ('claim_batch', '0006_capitationpayment'),
        ('core', '0019_extended_field'),
        ('product', '0006_insert_ceiling_type'),
        ('policy', '0004_add_medical_oficer_reading_rights'),
        ('medical', '0003_mutations'),
        ('claim', '0015_prohibit_interaction_for_enrolment_officer'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='claim',
            name='row_id',
        ),
        migrations.RemoveField(
            model_name='claimadmin',
            name='row_id',
        ),
        migrations.AddField(
            model_name='claim',
            name='adjuster',
            field=models.ForeignKey(blank=True, db_column='Adjuster', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='core.interactiveuser'),
        ),
        migrations.AddField(
            model_name='claim',
            name='admin',
            field=models.ForeignKey(blank=True, db_column='ClaimAdminId', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='claim.claimadmin'),
        ),
        migrations.AddField(
            model_name='claim',
            name='batch_run',
            field=models.ForeignKey(blank=True, db_column='RunID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='claim_batch.batchrun'),
        ),
        migrations.AddField(
            model_name='claim',
            name='feedback',
            field=models.OneToOneField(blank=True, db_column='FeedbackID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='claim.feedback'),
        ),
        migrations.AddField(
            model_name='claim',
            name='feedback_available',
            field=models.BooleanField(db_column='Feedback', default=False),
        ),
        migrations.AddField(
            model_name='claim',
            name='health_facility',
            field=models.ForeignKey(db_column='HFID', on_delete=django.db.models.deletion.DO_NOTHING, to='location.healthfacility'),
        ),
        migrations.AddField(
            model_name='claim',
            name='icd',
            field=models.ForeignKey(db_column='ICDID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_icds', to='medical.diagnosis'),
        ),
        migrations.AddField(
            model_name='claim',
            name='insuree',
            field=models.ForeignKey(db_column='InsureeID', on_delete=django.db.models.deletion.DO_NOTHING, to='insuree.insuree'),
        ),
        migrations.AddField(
            model_name='claim',
            name='json_ext',
            field=models.JSONField(blank=True, db_column='JsonExt', null=True),
        ),
        migrations.AddField(
            model_name='claim',
            name='uuid',
            field=models.CharField(db_column='ClaimUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AddField(
            model_name='claimadmin',
            name='has_login',
            field=models.BooleanField(blank=True, db_column='HasLogin', null=True),
        ),
        migrations.AddField(
            model_name='claimadmin',
            name='health_facility',
            field=models.ForeignKey(blank=True, db_column='HFId', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='location.healthfacility'),
        ),
        migrations.AddField(
            model_name='claimadmin',
            name='uuid',
            field=models.CharField(db_column='ClaimAdminUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AddField(
            model_name='claimdedrem',
            name='claim',
            field=models.ForeignKey(db_column='ClaimID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='dedrems', to='claim.claim'),
        ),
        migrations.AddField(
            model_name='claimdedrem',
            name='insuree',
            field=models.ForeignKey(blank=True, db_column='InsureeID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_ded_rems', to='insuree.insuree'),
        ),
        migrations.AddField(
            model_name='claimdedrem',
            name='policy',
            field=models.ForeignKey(blank=True, db_column='PolicyID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_ded_rems', to='policy.policy'),
        ),
        migrations.AddField(
            model_name='claimitem',
            name='claim',
            field=models.ForeignKey(db_column='ClaimID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='items', to='claim.claim'),
        ),
        migrations.AddField(
            model_name='claimitem',
            name='item',
            field=models.ForeignKey(db_column='ItemID', on_delete=django.db.models.deletion.DO_NOTHING, to='medical.item'),
        ),
        migrations.AddField(
            model_name='claimitem',
            name='json_ext',
            field=models.JSONField(blank=True, db_column='JsonExt', null=True),
        ),
        migrations.AddField(
            model_name='claimitem',
            name='policy',
            field=models.ForeignKey(blank=True, db_column='PolicyID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='policy.policy'),
        ),
        migrations.AddField(
            model_name='claimitem',
            name='product',
            field=models.ForeignKey(blank=True, db_column='ProdID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_items', to='product.product'),
        ),
        migrations.AddField(
            model_name='claimservice',
            name='claim',
            field=models.ForeignKey(db_column='ClaimID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='services', to='claim.claim'),
        ),
        migrations.AddField(
            model_name='claimservice',
            name='json_ext',
            field=models.JSONField(blank=True, db_column='JsonExt', null=True),
        ),
        migrations.AddField(
            model_name='claimservice',
            name='policy',
            field=models.ForeignKey(blank=True, db_column='PolicyID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='policy.policy'),
        ),
        migrations.AddField(
            model_name='claimservice',
            name='product',
            field=models.ForeignKey(blank=True, db_column='ProdID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_services', to='product.product'),
        ),
        migrations.AddField(
            model_name='claimservice',
            name='service',
            field=models.ForeignKey(db_column='ServiceID', on_delete=django.db.models.deletion.DO_NOTHING, to='medical.service'),
        ),
        migrations.AddField(
            model_name='feedback',
            name='claim',
            field=models.OneToOneField(blank=True, db_column='ClaimID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='claim.claim'),
        ),
        migrations.AddField(
            model_name='feedback',
            name='uuid',
            field=models.CharField(db_column='FeedbackUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AddField(
            model_name='feedbackprompt',
            name='claim_id',
            field=models.OneToOneField(blank=True, db_column='ClaimID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='claim.claim'),
        ),
        migrations.AlterField(
            model_name='claim',
            name='approval_status',
            field=models.SmallIntegerField(blank=True, db_column='ApprovalStatus', default=1, null=True),
        ),
        migrations.AlterField(
            model_name='claim',
            name='code',
            field=models.CharField(db_column='ClaimCode', max_length=8, unique=True),
        ),
        migrations.AlterField(
            model_name='claim',
            name='feedback_status',
            field=models.SmallIntegerField(blank=True, db_column='FeedbackStatus', default=1, null=True),
        ),
        migrations.AlterField(
            model_name='claim',
            name='icd_1',
            field=models.ForeignKey(blank=True, db_column='ICDID1', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_icd1s', to='medical.diagnosis'),
        ),
        migrations.AlterField(
            model_name='claim',
            name='icd_2',
            field=models.ForeignKey(blank=True, db_column='ICDID2', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_icd2s', to='medical.diagnosis'),
        ),
        migrations.AlterField(
            model_name='claim',
            name='icd_3',
            field=models.ForeignKey(blank=True, db_column='ICDID3', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_icd3s', to='medical.diagnosis'),
        ),
        migrations.AlterField(
            model_name='claim',
            name='icd_4',
            field=models.ForeignKey(blank=True, db_column='ICDID4', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='claim_icd4s', to='medical.diagnosis'),
        ),
        migrations.AlterField(
            model_name='claim',
            name='rejection_reason',
            field=models.SmallIntegerField(blank=True, db_column='RejectionReason', default=0, null=True),
        ),
        migrations.AlterField(
            model_name='claim',
            name='review_status',
            field=models.SmallIntegerField(blank=True, db_column='ReviewStatus', default=1, null=True),
        ),
        migrations.AlterField(
            model_name='claim',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='claimadmin',
            name='code',
            field=models.CharField(blank=True, db_column='ClaimAdminCode', max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='claimadmin',
            name='legacy_id',
            field=models.IntegerField(blank=True, db_column='LegacyID', null=True),
        ),
        migrations.AlterField(
            model_name='claimadmin',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='claimitem',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='claimservice',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='feedback',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
    ]



# Generated by Django 2.1.8 on 2019-04-03 08:37

import core.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Claim',
            fields=[
                ('id', models.AutoField(db_column='ClaimID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('category', models.CharField(blank=True, db_column='ClaimCategory', max_length=1, null=True)),
                ('code', models.CharField(db_column='ClaimCode', max_length=8)),
                ('date_from', core.fields.DateField(db_column='DateFrom')),
                ('date_to', core.fields.DateField(blank=True, db_column='DateTo', null=True)),
                ('status', models.SmallIntegerField(db_column='ClaimStatus')),
                ('adjustment', models.TextField(blank=True, db_column='Adjustment', null=True)),
                ('claimed', models.DecimalField(blank=True, db_column='Claimed', decimal_places=2, max_digits=18, null=True)),
                ('approved', models.DecimalField(blank=True, db_column='Approved', decimal_places=2, max_digits=18, null=True)),
                ('reinsured', models.DecimalField(blank=True, db_column='Reinsured', decimal_places=2, max_digits=18, null=True)),
                ('valuated', models.DecimalField(blank=True, db_column='Valuated', decimal_places=2, max_digits=18, null=True)),
                ('date_claimed', core.fields.DateField(db_column='DateClaimed')),
                ('date_processed', core.fields.DateField(blank=True, db_column='DateProcessed', null=True)),
                ('explanation', models.TextField(blank=True, db_column='Explanation', null=True)),
                ('feedback_status', models.SmallIntegerField(blank=True, db_column='FeedbackStatus', null=True)),
                ('review_status', models.SmallIntegerField(blank=True, db_column='ReviewStatus', null=True)),
                ('approval_status', models.SmallIntegerField(blank=True, db_column='ApprovalStatus', null=True)),
                ('rejection_reason', models.SmallIntegerField(blank=True, db_column='RejectionReason', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('validity_from_review', core.fields.DateTimeField(blank=True, db_column='ValidityFromReview', null=True)),
                ('validity_to_review', core.fields.DateTimeField(blank=True, db_column='ValidityToReview', null=True)),
                ('submit_stamp', core.fields.DateTimeField(blank=True, db_column='SubmitStamp', null=True)),
                ('process_stamp', core.fields.DateTimeField(blank=True, db_column='ProcessStamp', null=True)),
                ('remunerated', models.DecimalField(blank=True, db_column='Remunerated', decimal_places=2, max_digits=18, null=True)),
                ('guarantee_id', models.CharField(blank=True, db_column='GuaranteeId', max_length=50, null=True)),
                ('icd_1', models.IntegerField(blank=True, db_column='ICDID1', null=True)),
                ('icd_2', models.IntegerField(blank=True, db_column='ICDID2', null=True)),
                ('icd_3', models.IntegerField(blank=True, db_column='ICDID3', null=True)),
                ('icd_4', models.IntegerField(blank=True, db_column='ICDID4', null=True)),
                ('visit_type', models.CharField(blank=True, db_column='VisitType', max_length=1, null=True)),
                ('audit_user_id_review', models.IntegerField(blank=True, db_column='AuditUserIDReview', null=True)),
                ('audit_user_id_submit', models.IntegerField(blank=True, db_column='AuditUserIDSubmit', null=True)),
                ('audit_user_id_process', models.IntegerField(blank=True, db_column='AuditUserIDProcess', null=True)),
                ('row_id', models.BinaryField(blank=True, db_column='RowID', null=True)),
            ],
            options={
                'db_table': 'tblClaim',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='ClaimAdmin',
            fields=[
                ('id', models.AutoField(db_column='ClaimAdminId', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyId', null=True)),
                ('code', models.CharField(blank=True, db_column='ClaimAdminCode', max_length=8, null=True)),
                ('last_name', models.CharField(blank=True, db_column='LastName', max_length=100, null=True)),
                ('other_names', models.CharField(blank=True, db_column='OtherNames', max_length=100, null=True)),
                ('dob', models.DateField(blank=True, db_column='DOB', null=True)),
                ('email_id', models.CharField(blank=True, db_column='EmailId', max_length=200, null=True)),
                ('phone', models.CharField(blank=True, db_column='Phone', max_length=50, null=True)),
                ('validity_from', core.fields.DateTimeField(blank=True, db_column='ValidityFrom', null=True)),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AuditUserId', null=True)),
                ('row_id', models.BinaryField(blank=True, db_column='RowId', null=True)),
            ],
            options={
                'db_table': 'tblClaimAdmin',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='ClaimDiagnosisCode',
            fields=[
                ('id', models.AutoField(db_column='ICDID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('code', models.CharField(db_column='ICDCode', max_length=255)),
                ('name', models.CharField(db_column='ICDName', max_length=255)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('row_id', models.BinaryField(blank=True, db_column='RowID', null=True)),
            ],
            options={
                'db_table': 'tblICDCodes',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='ClaimItem',
            fields=[
                ('id', models.AutoField(db_column='ClaimItemID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('status', models.SmallIntegerField(db_column='ClaimItemStatus')),
                ('availability', models.BooleanField(db_column='Availability')),
                ('qty_provided', models.DecimalField(db_column='QtyProvided', decimal_places=2, max_digits=18)),
                ('qty_approved', models.DecimalField(blank=True, db_column='QtyApproved', decimal_places=2, max_digits=18, null=True)),
                ('price_asked', models.DecimalField(db_column='PriceAsked', decimal_places=2, max_digits=18)),
                ('price_adjusted', models.DecimalField(blank=True, db_column='PriceAdjusted', decimal_places=2, max_digits=18, null=True)),
                ('price_approved', models.DecimalField(blank=True, db_column='PriceApproved', decimal_places=2, max_digits=18, null=True)),
                ('price_valuated', models.DecimalField(blank=True, db_column='PriceValuated', decimal_places=2, max_digits=18, null=True)),
                ('explanation', models.TextField(blank=True, db_column='Explanation', null=True)),
                ('justification', models.TextField(blank=True, db_column='Justification', null=True)),
                ('rejection_reason', models.SmallIntegerField(blank=True, db_column='RejectionReason', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('validity_from_review', core.fields.DateTimeField(blank=True, db_column='ValidityFromReview', null=True)),
                ('validity_to_review', core.fields.DateTimeField(blank=True, db_column='ValidityToReview', null=True)),
                ('audit_user_id_review', models.IntegerField(blank=True, db_column='AuditUserIDReview', null=True)),
                ('limitation_value', models.DecimalField(blank=True, db_column='LimitationValue', decimal_places=2, max_digits=18, null=True)),
                ('limitation', models.CharField(blank=True, db_column='Limitation', max_length=1, null=True)),
                ('remunerated_amount', models.DecimalField(blank=True, db_column='RemuneratedAmount', decimal_places=2, max_digits=18, null=True)),
                ('deductable_amount', models.DecimalField(blank=True, db_column='DeductableAmount', decimal_places=2, max_digits=18, null=True)),
                ('exceed_ceiling_amount', models.DecimalField(blank=True, db_column='ExceedCeilingAmount', decimal_places=2, max_digits=18, null=True)),
                ('price_origin', models.CharField(blank=True, db_column='PriceOrigin', max_length=1, null=True)),
                ('exceed_ceiling_amount_category', models.DecimalField(blank=True, db_column='ExceedCeilingAmountCategory', decimal_places=2, max_digits=18, null=True)),
            ],
            options={
                'db_table': 'tblClaimItems',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='ClaimService',
            fields=[
                ('id', models.AutoField(db_column='ClaimServiceID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('status', models.SmallIntegerField(db_column='ClaimServiceStatus')),
                ('qty_provided', models.DecimalField(db_column='QtyProvided', decimal_places=2, max_digits=18)),
                ('qty_approved', models.DecimalField(blank=True, db_column='QtyApproved', decimal_places=2, max_digits=18, null=True)),
                ('price_asked', models.DecimalField(db_column='PriceAsked', decimal_places=2, max_digits=18)),
                ('price_adjusted', models.DecimalField(blank=True, db_column='PriceAdjusted', decimal_places=2, max_digits=18, null=True)),
                ('price_approved', models.DecimalField(blank=True, db_column='PriceApproved', decimal_places=2, max_digits=18, null=True)),
                ('price_valuated', models.DecimalField(blank=True, db_column='PriceValuated', decimal_places=2, max_digits=18, null=True)),
                ('explanation', models.TextField(blank=True, db_column='Explanation', null=True)),
                ('justification', models.TextField(blank=True, db_column='Justification', null=True)),
                ('rejection_reason', models.SmallIntegerField(blank=True, db_column='RejectionReason', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('validity_from_review', core.fields.DateTimeField(blank=True, db_column='ValidityFromReview', null=True)),
                ('validity_to_review', core.fields.DateTimeField(blank=True, db_column='ValidityToReview', null=True)),
                ('audit_user_id_review', models.IntegerField(blank=True, db_column='AuditUserIDReview', null=True)),
                ('limitation_value', models.DecimalField(blank=True, db_column='LimitationValue', decimal_places=2, max_digits=18, null=True)),
                ('limitation', models.CharField(blank=True, db_column='Limitation', max_length=1, null=True)),
                ('remunerated_amount', models.DecimalField(blank=True, db_column='RemuneratedAmount', decimal_places=2, max_digits=18, null=True)),
                ('deductable_amount', models.DecimalField(blank=True, db_column='DeductableAmount', decimal_places=2, max_digits=18, null=True)),
                ('exceed_ceiling_amount', models.DecimalField(blank=True, db_column='ExceedCeilingAmount', decimal_places=2, max_digits=18, null=True)),
                ('price_origin', models.CharField(blank=True, db_column='PriceOrigin', max_length=1, null=True)),
                ('exceed_ceiling_amount_category', models.DecimalField(blank=True, db_column='ExceedCeilingAmountCategory', decimal_places=2, max_digits=18, null=True)),
            ],
            options={
                'db_table': 'tblClaimServices',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='Feedback',
            fields=[
                ('id', models.AutoField(db_column='FeedbackID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('care_rendered', models.BooleanField(blank=True, db_column='CareRendered', null=True)),
                ('payment_asked', models.BooleanField(blank=True, db_column='PaymentAsked', null=True)),
                ('drug_prescribed', models.BooleanField(blank=True, db_column='DrugPrescribed', null=True)),
                ('drug_received', models.BooleanField(blank=True, db_column='DrugReceived', null=True)),
                ('asessment', models.SmallIntegerField(blank=True, db_column='Asessment', null=True)),
                ('officer_id', models.IntegerField(blank=True, db_column='CHFOfficerCode', null=True)),
                ('feedback_date', core.fields.DateTimeField(blank=True, db_column='FeedbackDate', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
            ],
            options={
                'db_table': 'tblFeedback',
                'managed': False,
            },
        ),
    ]



from claim.services import update_claims_dedrems, set_claims_status
from claim.models import Claim, ClaimDedRem, ClaimItem, ClaimDetail, ClaimService, ClaimServiceItem, ClaimServiceService
from claim.test_helpers import create_test_claim, create_test_claimservice, create_test_claimitem, \
    mark_test_claim_as_processed, delete_claim_with_itemsvc_dedrem_and_history
from core.test_helpers import create_test_officer

from claim.validations import get_claim_category, validate_claim, validate_assign_prod_to_claimitems_and_services, \
    process_dedrem, REJECTION_REASON_WAITING_PERIOD_FAIL, REJECTION_REASON_INVALID_ITEM_OR_SERVICE
from core.models import User, InteractiveUser
from django.test import TestCase
from insuree.models import Family, Insuree
from insuree.test_helpers import create_test_insuree
from location.models import HealthFacility
from medical_pricelist.test_helpers import add_service_to_hf_pricelist, add_item_to_hf_pricelist
from medical.models import ServiceItem, ServiceService
from product.models import ProductItemOrService
from medical.test_helpers import create_test_service, create_test_item
from medical_pricelist.test_helpers import add_service_to_hf_pricelist, add_item_to_hf_pricelist, \
    update_pricelist_service_detail_in_hf_pricelist, update_pricelist_item_detail_in_hf_pricelist
from policy.test_helpers import create_test_policy
 

# default arguments should not pass a list or a dict because they're mutable but we don't risk mutating them here:
# noinspection PyDefaultArgument,DuplicatedCode
from product.test_helpers import create_test_product, create_test_product_service, create_test_product_item


class ValidationTest(TestCase):
    service_H = None
    service_O = None
    service_D = None
    service_A = None
    service_A_invalid = None

    def setUp(self) -> None:
        super(ValidationTest, self).setUp()
        self.i_user = InteractiveUser(
            login_name="test_batch_run", audit_user_id=978911, id=97891
        )
        self.user = User(i_user=self.i_user)

        self.service_H = create_test_service("H")
        self.service_O = create_test_service("O")
        self.service_D = create_test_service("D")
        self.service_A = create_test_service("A")
        self.service_A_invalid = create_test_service("A", False)

        self.item_1 = create_test_item("D")

    def test_get_claim_category_S(self):
        # Given
        claim = create_test_claim()
        service1 = create_test_claimservice(claim, "S")

        # when
        category = get_claim_category(claim)

        # then
        self.assertIsNotNone(category)
        self.assertEquals(category, "S")

        # tearDown
        service1.delete()
        claim.delete()

    def test_get_claim_category_D(self):
        # Given
        claim = create_test_claim()
        service1 = create_test_claimservice(claim, "D")

        # when
        category = get_claim_category(claim)

        # then
        self.assertIsNotNone(category)
        self.assertEquals(category, "D")

        # tearDown
        service1.delete()
        claim.delete()

    def test_get_claim_category_mix(self):
        # Given
        claim = create_test_claim()
        services = [
            create_test_claimservice(claim, "H"),
            create_test_claimservice(claim, "O"),
            create_test_claimservice(claim, None),
            create_test_claimservice(claim, "A"),
            create_test_claimservice(claim, "S"),
        ]

        # when
        category = get_claim_category(claim)

        # then
        self.assertIsNotNone(category)
        self.assertEquals(category, "S")

        # tearDown
        for service in services:
            service.delete()
        claim.delete()

    def test_get_claim_category_some_invalid(self):
        # Given
        claim = create_test_claim()
        services = [
            create_test_claimservice(claim, "H", False),
            create_test_claimservice(claim, "A", False),
            create_test_claimservice(claim, "S"),
        ]

        # when
        category = get_claim_category(claim)

        # then
        self.assertIsNotNone(category)
        self.assertEquals(category, "S")

        # tearDown
        for service in services:
            service.delete()
        claim.delete()

    def test_get_claim_category_null(self):
        # Given
        claim = create_test_claim()
        service1 = create_test_claimservice(claim, None)

        # when
        category = get_claim_category(claim)

        # then
        self.assertIsNotNone(category)
        self.assertEquals(category, "V")

        # tearDown
        service1.delete()
        claim.delete()

    # This test cannot be performed because the database constraints don't allow a null date_from.
    # def test_validate_claim_target_date(self):
    #     # Given
    #     claim = create_test_claim(custom_props={"date_from": None, "date_to": None})
    #     service1 = create_test_claimservice(claim, "S")
    #
    #     # When
    #     errors = validate_claim(claim)
    #
    #     # Then
    #     self.assertEquals(len(errors), 1, "The claim should fail the target date validation")
    #     self.assertEquals(errors[0].code, 9, "The claim should fail the target date validation with code 9")
    #     self.assertTrue("date" in errors[0].message.lower())
    #
    #     # tearDown
    #     service1.delete()
    #     claim.delete()

    def test_validate_pricelist_hf1(self):
        # When the claimitem points to a pricelist that doesn't correspond to the claim HF
        hf_without_pricelist = HealthFacility.objects.filter(items_pricelist__id__isnull=True).first()
        self.assertIsNotNone(hf_without_pricelist, "This test requires a health facility without a price list item")
        # Given
        claim = create_test_claim({"health_facility_id": hf_without_pricelist.id})
        service1 = create_test_claimservice(claim, "S")
        item1 = create_test_claimitem(claim, "D")

        # When
        errors = validate_claim(claim, True)

        # Then
        claim.refresh_from_db()
        service1.refresh_from_db()
        item1.refresh_from_db()
        self.assertGreaterEqual(len(errors), 1, "Should raise at least one error")
        error1 = [e for e in errors if e['code'] == 1]  # all services rejected
        self.assertGreaterEqual(len(error1), 1, "There should be an error code 1")
        self.assertEquals(item1.rejection_reason, 2, "Database was updated with rejection reason")

        # tearDown
        service1.delete()
        item1.delete()
        claim.delete()

    def test_validate_family(self):
        # When the insuree family is invalid
        # Given
        invalid_insuree = Insuree.objects.filter(family__in=Family.objects.filter(validity_to__isnull=False)).first()
        self.assertIsNotNone(invalid_insuree)
        claim = create_test_claim({"insuree_id": invalid_insuree.id})
        service1 = create_test_claimservice(claim, "S")
        item1 = create_test_claimitem(claim, "D")

        # When
        errors = validate_claim(claim, True)

        # Then
        claim.refresh_from_db()
        item1.refresh_from_db()
        self.assertGreaterEqual(len(errors), 1, "Should raise at least one error")
        error7 = [e for e in errors if e['code'] == 7]
        self.assertGreaterEqual(len(error7), 1, "There should be 1 error code 7: invalid insuree")
        self.assertEquals(item1.rejection_reason, 7, "Database was updated with rejection reason")

        # tearDown
        service1.delete()
        item1.delete()
        claim.delete()

    def test_validate_max_visits(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("VISIT", custom_props={"max_no_visits": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V")
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a visit should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        self.assertEquals(len(errors), 0, "The first visit should be accepted")

        # a second visit should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, "V")
        errors = validate_claim(claim2, True)
        # TODO Temporarily disabled
        # self.assertGreater(len(errors), 0, "The second visit should be refused")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_validate_patient_category(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("VISIT", custom_props={"max_no_visits": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V", custom_props={"patient_category": 1})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # The insuree has a patient_category of 6, not matching the service category
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)

        # Then
        claim1.refresh_from_db()
        self.assertEquals(len(errors), 2)
        self.assertEquals(errors[0]['code'], 1)  # claimed rejected because all services are rejected
        self.assertEquals(claim1.services.first().rejection_reason, 4)  # reason is wrong insuree mask

        # tearDown
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_frequency(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("CSECT")
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("C", custom_props={"code": "G34B", "frequency": 180})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a visit should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first visit should be accepted")

        # a second visit should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0, "The second service should be refused as it is withing 180 days")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_limit_no(self):
        # When the insuree already reaches his limit number
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("CSECT")
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("C", custom_props={"code": "G34B"})
        product_service = create_test_product_service(product, service, custom_props={"limit_no_adult": 1})
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a visit should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id, "qty_provided": 1})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first visit should be accepted")

        # a second visit should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0, "The second service should be refused")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_limit_delivery(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("DELIV", custom_props={"max_no_delivery": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("D", custom_props={"code": "G34C"})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a delivery should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first visit should be accepted")

        # a second delivery should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0, "The second service should fail because there is already one delivery")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_limit_hospital(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("DELIV", custom_props={"max_no_hospitalization": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("H", custom_props={"code": "HHHH"})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a delivery should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first hospitalization should be accepted")

        # a second delivery should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0,
                           "The second service should fail because there is already one hospitalization")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_limit_surgery(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("DELIV", custom_props={"max_no_surgery": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("S", custom_props={"code": "SSSS"})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a delivery should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first surgery should be accepted")

        # a second delivery should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0, "The second service should fail because there is already one surgery")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_limit_visit(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("DELIV", custom_props={"max_no_visits": 1})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V", custom_props={"code": "VVVV"})
        product_service = create_test_product_service(product, service)
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a delivery should be accepted
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)
        mark_test_claim_as_processed(claim1)

        self.assertEquals(len(errors), 0, "The first visit should be accepted")

        # a second delivery should be denied
        claim2 = create_test_claim({"insuree_id": insuree.id})
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertGreater(len(errors), 0, "The second service should fail because there is already one visit")

        # Then
        claim1.refresh_from_db()
        claim2.refresh_from_db()

        # tearDown
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_waiting_period(self):
        # When the insuree already reaches his limit of visits
        # Given
        from core import datetime
        insuree = create_test_insuree()
        child_insuree = create_test_insuree( with_family = False,
            custom_props={
            "dob": datetime.datetime(2020, 1, 1),
            "family": insuree.family
        })
        self.assertIsNotNone(insuree)
        product = create_test_product("CSECT")
        policy = create_test_policy(product, insuree, link=True)
        policy_child = create_test_policy(product, child_insuree, link=True)
        service = create_test_service("C")
        product_service = create_test_product_service(
            product, service, custom_props={"waiting_period_adult": 6, "waiting_period_child": 0})
        pricelist_detail = add_service_to_hf_pricelist(service)

        # A first claim for a visit within the waiting period should be refused
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(claim1, custom_props={"service_id": service.id})
        errors = validate_claim(claim1, True)

        self.assertEqual(len(errors), 2, "An adult visit within the waiting period should be refused")
        self.assertEqual(claim1.services.first().rejection_reason, REJECTION_REASON_WAITING_PERIOD_FAIL)

        # a visit after the waiting period should be fine
        claim2 = create_test_claim({
            "insuree_id": insuree.id,
            "date_from": datetime.datetime(2020, 2, 1),
            "date_to": datetime.datetime(2020, 2, 1),
            "date_claimed": datetime.datetime(2020, 2, 1),
        })
        service2 = create_test_claimservice(claim2, custom_props={"service_id": service.id})
        errors = validate_claim(claim2, True)
        self.assertEqual(len(errors), 0, "This one should be accepted as after the waiting period")

        # a child should not have the waiting period
        claim3 = create_test_claim({"insuree_id": child_insuree.id})
        service3 = create_test_claimservice(claim3, custom_props={"service_id": service.id})
        errors = validate_claim(claim3, True)
        self.assertEqual(len(errors), 0, "The child has no waiting period")

        # tearDown
        service3.delete()
        claim3.delete()
        service2.delete()
        claim2.delete()
        service1.delete()
        claim1.delete()
        policy.insuree_policies.all().delete()
        policy.delete()
        policy_child.insuree_policies.all().delete()
        policy_child.delete()
        product_service.delete()
        pricelist_detail.delete()
        service.delete()
        product.delete()

    def test_submit_claim_dedrem(self):
        # When the insuree already reaches his limit of visits
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("VISIT", custom_props={})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V", custom_props={})
        item = create_test_item("D", custom_props={})
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        item1 = create_test_claimitem(
            claim1, "D", custom_props={"item_id": item.id})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)

        # Then
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEqual(len(errors), 0)
        self.assertEqual(item1.price_adjusted, 100)
        self.assertEqual(item1.price_valuated, 700)
        self.assertEqual(item1.deductable_amount, 0)
        self.assertEqual(item1.exceed_ceiling_amount, 0)
        self.assertIsNone(item1.exceed_ceiling_amount_category)
        self.assertEqual(item1.remunerated_amount, 700)
        self.assertEqual(claim1.status, Claim.STATUS_VALUATED)
        self.assertEqual(claim1.audit_user_id_process, -1)
        self.assertIsNotNone(claim1.process_stamp)
        self.assertIsNotNone(claim1.date_processed)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        self.assertEqual(dedrem_qs.count(), 1)
        dedrem1 = dedrem_qs.first()
        self.assertEqual(dedrem1.policy_id, item1.policy_id)
        self.assertEqual(dedrem1.insuree_id, claim1.insuree_id)
        self.assertEqual(dedrem1.audit_user_id, -1)
        self.assertEqual(dedrem1.ded_g, 0)
        self.assertEqual(dedrem1.rem_g, 1400)
        self.assertEqual(dedrem1.rem_op, 1400)
        self.assertIsNone(dedrem1.rem_ip)
        self.assertEqual(dedrem1.rem_surgery, 0)
        self.assertEqual(dedrem1.rem_consult, 0)
        self.assertEqual(dedrem1.rem_hospitalization, 0)
        self.assertIsNotNone(claim1.validity_from)
        self.assertIsNone(claim1.validity_to)

        # tearDown
        dedrem_qs.delete()
        service1.delete()
        item1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_submit_claim_dedrem_limit_delivery(self):
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        service = create_test_service("D", custom_props={})
        item = create_test_item("D", custom_props={})

        product = create_test_product("BCUL0001", custom_props={
            "name": "Basic Cover Ultha",
            "lump_sum": 10_000,
            "max_amount_delivery": 55,
        })
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        policy = create_test_policy(product, insuree, link=True)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        # The insuree has a patient_category of 6, not matching the service category
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        item1 = create_test_claimitem(
            claim1, "D", custom_props={"item_id": item.id})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)

        # Then
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEqual(len(errors), 0)
        self.assertEqual(item1.price_adjusted, 100)
        self.assertEqual(item1.price_valuated, 55)
        self.assertEqual(item1.deductable_amount, 0)
        self.assertEqual(item1.exceed_ceiling_amount, 0)
        self.assertIsNone(item1.exceed_ceiling_amount_category)
        self.assertEqual(item1.remunerated_amount, 55)
        self.assertEqual(claim1.status, Claim.STATUS_VALUATED)
        self.assertEqual(claim1.audit_user_id_process, -1)
        self.assertIsNotNone(claim1.process_stamp)
        self.assertIsNotNone(claim1.date_processed)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        self.assertEqual(dedrem_qs.count(), 1)
        dedrem1 = dedrem_qs.first()
        self.assertEqual(dedrem1.policy_id, item1.policy_id)
        self.assertEqual(dedrem1.insuree_id, claim1.insuree_id)
        self.assertEqual(dedrem1.audit_user_id, -1)
        self.assertEqual(dedrem1.ded_g, 0)
        self.assertEqual(dedrem1.rem_g, 55)
        self.assertIsNotNone(claim1.validity_from)
        self.assertIsNone(claim1.validity_to)

        # tearDown
        dedrem_qs.delete()
        service1.delete()
        item1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_submit_claim_dedrem_limit_consultation(self):
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        service = create_test_service("C", custom_props={})
        item = create_test_item("C", custom_props={})

        product = create_test_product("BCUL0001", custom_props={
            "name": "Basic Cover Ultha",
            "lump_sum": 10_000,
            "max_amount_consultation": 55,
        })
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        policy = create_test_policy(product, insuree, link=True)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        # The insuree has a patient_category of 6, not matching the service category
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        item1 = create_test_claimitem(
            claim1, "C", custom_props={"item_id": item.id})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)

        # Then
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEqual(len(errors), 0)
        self.assertEqual(item1.price_adjusted, 100)
        self.assertEqual(item1.price_valuated, 55)
        self.assertEqual(item1.deductable_amount, 0)
        self.assertEqual(item1.exceed_ceiling_amount, 0)
        self.assertIsNone(item1.exceed_ceiling_amount_category)
        self.assertEqual(item1.remunerated_amount, 55)
        self.assertEqual(claim1.status, Claim.STATUS_VALUATED)
        self.assertEqual(claim1.audit_user_id_process, -1)
        self.assertIsNotNone(claim1.process_stamp)
        self.assertIsNotNone(claim1.date_processed)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        self.assertEqual(dedrem_qs.count(), 1)
        dedrem1 = dedrem_qs.first()
        self.assertEqual(dedrem1.policy_id, item1.policy_id)
        self.assertEqual(dedrem1.insuree_id, claim1.insuree_id)
        self.assertEqual(dedrem1.audit_user_id, -1)
        self.assertEqual(dedrem1.ded_g, 0)
        self.assertEqual(dedrem1.rem_g, 55)
        self.assertIsNotNone(claim1.validity_from)
        self.assertIsNone(claim1.validity_to)

        # tearDown
        dedrem_qs.delete()
        service1.delete()
        item1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_submit_claim_dedrem_limit_antenatal(self):
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        service = create_test_service("A", custom_props={})
        item = create_test_item("A", custom_props={})

        product = create_test_product("BCUL0001", custom_props={
            "name": "Basic Cover Ultha",
            "lump_sum": 10_000,
            "max_amount_antenatal": 55,
        })
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        policy = create_test_policy(product, insuree, link=True)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        # The insuree has a patient_category of 6, not matching the service category
        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        item1 = create_test_claimitem(
            claim1, "A", custom_props={"item_id": item.id})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)

        # Then
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEqual(len(errors), 0)
        self.assertEqual(item1.price_adjusted, 100)
        self.assertEqual(item1.price_valuated, 55)
        self.assertEqual(item1.deductable_amount, 0)
        self.assertEqual(item1.exceed_ceiling_amount, 0)
        self.assertIsNone(item1.exceed_ceiling_amount_category)
        self.assertEqual(item1.remunerated_amount, 55)
        self.assertEqual(claim1.status, Claim.STATUS_VALUATED)
        self.assertEqual(claim1.audit_user_id_process, -1)
        self.assertIsNotNone(claim1.process_stamp)
        self.assertIsNotNone(claim1.date_processed)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        self.assertEqual(dedrem_qs.count(), 1)
        dedrem1 = dedrem_qs.first()
        self.assertEqual(dedrem1.policy_id, item1.policy_id)
        self.assertEqual(dedrem1.insuree_id, claim1.insuree_id)
        self.assertEqual(dedrem1.audit_user_id, -1)
        self.assertEqual(dedrem1.ded_g, 0)
        self.assertEqual(dedrem1.rem_g, 55)
        self.assertIsNotNone(claim1.validity_from)
        self.assertIsNone(claim1.validity_to)

        # tearDown
        dedrem_qs.delete()
        service1.delete()
        item1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_review_reject_update_dedrem(self):
        """
        This test creates a claim, submits it so that it gets dedrem entries,
        then submits a review rejecting part of it, then process the claim.
        It should not be processed (which was ok) but the dedrem should be deleted.
        """
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        service = create_test_service("A", custom_props={"name": "test_review_reject_delete_dedrem"})
        item = create_test_item("A", custom_props={"name": "test_review_reject_delete_dedrem"})

        product = create_test_product("BCUL0001", custom_props={
            "name": "Basic Cover Ultha deldedrem",
            "lump_sum": 10_000,
        })
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        policy = create_test_policy(product, insuree, link=True)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id, "qty_provided": 2})
        item1 = create_test_claimitem(
            claim1, "A", custom_props={"item_id": item.id, "qty_provided": 3})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, False)

        self.assertEqual(len(errors), 0)
        # Make sure that the dedrem was generated
        dedrem = ClaimDedRem.objects.filter(claim=claim1).first()
        self.assertIsNotNone(dedrem)
        self.assertEquals(dedrem.rem_g, 500)  # 100*2 + 100*3

        # Review the claim and reject all of it
        # A partial rejection would still trigger the process_dedrem and be fine
        item1.qty_approved = 1
        item1.price_approved = 37
        item1.status = ClaimItem.STATUS_PASSED
        item1.audit_user_id_review = -1
        item1.justification = "Review comment item"
        item1.save()

        service1.qty_approved = 1
        service1.price_approved = 53
        service1.status = ClaimItem.STATUS_PASSED
        service1.audit_user_id_review = -1
        service1.justification = "Review comment svc"
        service1.save()

        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()

        set_claims_status([claim1.uuid], "review_status", Claim.REVIEW_DELIVERED)
        update_claims_dedrems([claim1.uuid], self.user)

        # Then dedrem should have been updated
        dedrem = ClaimDedRem.objects.filter(claim=claim1).first()
        self.assertIsNotNone(dedrem)
        self.assertEquals(dedrem.rem_g, 200)  # 100*1 + 100*1

        # tearDown
        # dedrem.delete() # already done if the test passed
        delete_claim_with_itemsvc_dedrem_and_history(claim1)
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_review_reject_delete_dedrem(self):
        """
        This test creates a claim, submits it so that it gets dedrem entries,
        then submits a review rejecting part of it, then process the claim.
        It should not be processed (which was ok) but the dedrem should be deleted.
        """
        # Given
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        service = create_test_service("A", custom_props={"name": "test_review_reject_delete_dedrem"})
        item = create_test_item("A", custom_props={"name": "test_review_reject_delete_dedrem"})

        product = create_test_product("BCUL0001", custom_props={
            "name": "Basic Cover Ultha deldedrem",
            "lump_sum": 10_000,
        })
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        policy = create_test_policy(product, insuree, link=True)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id, "qty_provided": 2})
        item1 = create_test_claimitem(
            claim1, "A", custom_props={"item_id": item.id, "qty_provided": 3})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, False)

        self.assertEqual(len(errors), 0)
        # Make sure that the dedrem was generated
        dedrem = ClaimDedRem.objects.filter(claim=claim1).first()
        self.assertIsNotNone(dedrem)
        self.assertEquals(dedrem.rem_g, 500)  # 100*2 + 100*3

        # Review the claim and reject all of it
        # A partial rejection would still trigger the process_dedrem and be fine
        item1.qty_approved = 0
        item1.price_approved = 0
        item1.status = ClaimItem.STATUS_REJECTED
        item1.rejection_reason = -1
        item1.audit_user_id_review = -1
        item1.justification = "Review comment item"
        item1.save()

        service1.qty_approved = 0
        service1.price_approved = 0
        service1.status = ClaimService.STATUS_REJECTED
        service1.rejection_reason = -1
        service1.audit_user_id_review = -1
        service1.justification = "Review comment svc"
        service1.save()

        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()

        set_claims_status([claim1.uuid], "review_status", Claim.REVIEW_DELIVERED)
        update_claims_dedrems([claim1.uuid], self.user)

        errors = validate_claim(claim1, True)
        if len(errors) == 0:
            errors += validate_assign_prod_to_claimitems_and_services(claim1)
            errors += process_dedrem(claim1, -1, False)

        # The claim should be globally rejected since the review rejected all items/svc
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEquals(claim1.status, Claim.STATUS_REJECTED)
        self.assertEquals(claim1.rejection_reason, REJECTION_REASON_INVALID_ITEM_OR_SERVICE)
        self.assertEquals(item1.status, ClaimDetail.STATUS_REJECTED)
        self.assertEquals(item1.rejection_reason, -1)
        self.assertEquals(service1.status, ClaimDetail.STATUS_REJECTED)
        self.assertEquals(service1.rejection_reason, -1)

        # Then dedrem should have been deleted
        dedrem = ClaimDedRem.objects.filter(claim=claim1).first()
        self.assertIsNone(dedrem)

        # tearDown
        # dedrem.delete() # already done if the test passed
        delete_claim_with_itemsvc_dedrem_and_history(claim1)
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_submit_claim_dedrem_update_pricelist_detail(self):
        '''
        This test replicates the functionality of test_submit_claim_dedrem,
        with the additional step of updating items and services prior to dedrem calculation.
        Despite the updates, the results should remain unaffected as the prices
        for these services/items are sourced from the pricelist detail's state at the time of
        coalse(claim.dateto, claim.datefrom).
        '''
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("VISIT", custom_props={})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V", custom_props={})
        item = create_test_item("D", custom_props={})
        product_service = create_test_product_service(product, service)
        product_item = create_test_product_item(product, item)
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        claim1 = create_test_claim({"insuree_id": insuree.id})
        service1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        item1 = create_test_claimitem(
            claim1, "D", custom_props={"item_id": item.id})
        errors = validate_claim(claim1, True)
        errors += validate_assign_prod_to_claimitems_and_services(claim1)
        update_pricelist_service_detail_in_hf_pricelist(pricelist_detail1, custom_props={"price_overrule": 21})
        update_pricelist_item_detail_in_hf_pricelist(pricelist_detail1, custom_props={"price_overrule": 37})
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)

        # Then
        claim1.refresh_from_db()
        item1.refresh_from_db()
        service1.refresh_from_db()
        self.assertEqual(len(errors), 0)
        self.assertEqual(item1.price_adjusted, 100)
        self.assertEqual(item1.price_valuated, 700)
        self.assertEqual(item1.deductable_amount, 0)
        self.assertEqual(item1.exceed_ceiling_amount, 0)
        self.assertIsNone(item1.exceed_ceiling_amount_category)
        self.assertEqual(item1.remunerated_amount, 700)
        self.assertEqual(claim1.status, Claim.STATUS_VALUATED)
        self.assertEqual(claim1.audit_user_id_process, -1)
        self.assertIsNotNone(claim1.process_stamp)
        self.assertIsNotNone(claim1.date_processed)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        self.assertEqual(dedrem_qs.count(), 1)
        dedrem1 = dedrem_qs.first()
        self.assertEqual(dedrem1.policy_id, item1.policy_id)
        self.assertEqual(dedrem1.insuree_id, claim1.insuree_id)
        self.assertEqual(dedrem1.audit_user_id, -1)
        self.assertEqual(dedrem1.ded_g, 0)
        self.assertEqual(dedrem1.rem_g, 1400)
        self.assertEqual(dedrem1.rem_op, 1400)
        self.assertIsNone(dedrem1.rem_ip)
        self.assertEqual(dedrem1.rem_surgery, 0)
        self.assertEqual(dedrem1.rem_consult, 0)
        self.assertEqual(dedrem1.rem_hospitalization, 0)
        self.assertIsNotNone(claim1.validity_from)
        self.assertIsNone(claim1.validity_to)

        # tearDown
        dedrem_qs.delete()
        service1.delete()
        item1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()

    def test_set_status(self):
        class DummyUser:
            id_for_audit=-1
        insuree = create_test_insuree()
        officer = create_test_officer(villages=[insuree.current_village or insuree.family.location])
        claim = create_test_claim(custom_props={'status':Claim.STATUS_CHECKED, 
                                                'insuree':insuree})
        restult =set_claims_status([claim.uuid], 'feedback_status', Claim.FEEDBACK_SELECTED, user = DummyUser())
        claim.refresh_from_db()
        self.assertEqual(claim.feedback_status, Claim.FEEDBACK_SELECTED)
    
    def test_submit_claim_with_different_packatypes(self):
        from claim.apps import ClaimConfig
        ClaimConfig.native_code_for_services=False
        insuree = create_test_insuree()
        self.assertIsNotNone(insuree)
        product = create_test_product("VISIT", custom_props={})
        policy = create_test_policy(product, insuree, link=True)
        service = create_test_service("V", custom_props={"packagetype": "F"})
        item = create_test_item("D", custom_props={})
        product_service = create_test_product_service(product, service)
        product_service.price_origin = ProductItemOrService.ORIGIN_CLAIM
        product_service.save()
        product_item = create_test_product_item(product, item)
        product_item.price_origin = ProductItemOrService.ORIGIN_CLAIM
        product_item.save()
        pricelist_detail1 = add_service_to_hf_pricelist(service)
        pricelist_detail2 = add_item_to_hf_pricelist(item)

        claim1 = create_test_claim({"insuree_id": insuree.id})
        claimservice1 = create_test_claimservice(
            claim1, custom_props={"service_id": service.id})
        clalimitem1 = create_test_claimitem(
            claim1, "D", custom_props={"item_id": item.id})
        # Set the price_adjusted to 4000
        claimservice1.price_adjusted = 4000
        claimservice1.save()
        # set the service price to 1000 lower than the price_adjusted
        service.price = 1000
        service.save()
        service.refresh_from_db()
        errors = validate_claim(claim1, True)
        errors = validate_assign_prod_to_claimitems_and_services(claim1)
        errors += process_dedrem(claim1, -1, True)
        self.assertEqual(len(errors), 0)
        # The claimservice1's price_adjusted should be the service price
        # because the price_adjusted is greater than the service (4000 > 1000)
        claimservice1.refresh_from_db()
        self.assertEqual(claimservice1.price_adjusted, 1000)

        # Set the price_adjusted to None
        claimservice1.price_adjusted = None
        claimservice1.price_asked = 6000
        service.price = 750
        service.save()
        claimservice1.save()
        validate_claim(claim1, True)
        validate_assign_prod_to_claimitems_and_services(claim1)
        process_dedrem(claim1, -1, True)
        claimservice1.refresh_from_db()
        service.refresh_from_db()
        # The claimservice1's price_adjusted should also be the service's price
        # because the price_asked is now also greater than the service (6000 > 750)
        self.assertEqual(claimservice1.price_adjusted, 750)

        # Change the price_origin
        product_service.price_origin = ProductItemOrService.ORIGIN_PRICELIST
        product_service.save()
        product_item.price_origin = ProductItemOrService.ORIGIN_PRICELIST
        product_item.save()
        service.packagetype = "P"
        service.save()
        # serviceservice = ServiceService.objects.create(
        #     servicelinkedService=service,
        #     service_id = claimservice1.service_id,
        #     price_asked = 3,
        #     qty_provided = 2
        # )
        claimserviceservice = ClaimServiceService.objects.create(
            service = service,
            claim_service = claimservice1,
            qty_displayed = 5,
            qty_provided = 4,
            price_asked = 300,
        )
        claimserviceitem = ClaimServiceItem.objects.create(
            item = item,
            claim_service = claimservice1,
            qty_displayed = 2,
            qty_provided = 3,
            price_asked = 500,
        )
        serviceitem = ServiceItem.objects.create(
            servicelinkedItem=service,
            item = clalimitem1.item,
            price_asked = 12,
            qty_provided = 11
        )

        validate_claim(claim1, True)
        validate_assign_prod_to_claimitems_and_services(claim1)
        process_dedrem(claim1, -1, True)
        claimservice1.refresh_from_db()
        # claimservicesitem.qty_provided 2 is not equal to qty_displayed on the claimserviceservice
        # so the price_adjusted is set to 0
        self.assertEqual(claimservice1.price_adjusted, 0)

        claimserviceservice.qty_displayed = 2
        claimserviceservice.save()
        claimserviceservice.refresh_from_db
        validate_claim(claim1, True)
        validate_assign_prod_to_claimitems_and_services(claim1)
        process_dedrem(claim1, -1, True)
        claimservice1.refresh_from_db()
        # service item.qty_provided 11 is not equal to qty_displayed on the claimserviceitem
        # so the price_adjusted is set to 0
        self.assertEqual(claimservice1.price_adjusted, 0)

        dedrem_qs = ClaimDedRem.objects.filter(claim=claim1)
        # tearDown
        dedrem_qs.delete()
        # serviceservice.delete()
        claimserviceservice.delete()
        claimserviceitem.delete()
        claimservice1.delete()
        serviceitem.delete()
        clalimitem1.delete()
        claim1.delete()
        policy.insuree_policies.first().delete()
        policy.delete()
        product_item.delete()
        product_service.delete()
        pricelist_detail1.delete()
        pricelist_detail2.delete()
        service.delete()
        item.delete()
        product.delete()
        ClaimConfig.native_code_for_services=True


# Generated by Django 3.2.18 on 2023-06-02 11:50
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    psql_code = 'select 1'

    dependencies = [
        ('claim', '0017_set_managed_to_true'),
    ]

    operations = [
        migrations.RunSQL('ALTER TABLE [tblClaim] ALTER COLUMN [JsonExt] NVARCHAR(MAX)'
                          if settings.MSSQL else psql_code,
                          reverse_sql='ALTER TABLE [tblClaim] ALTER COLUMN [JsonExt] TEXT'
                          if settings.MSSQL else psql_code),
        migrations.RunSQL('ALTER TABLE [tblClaimServices] ALTER COLUMN [JsonExt] NVARCHAR(MAX)'
                          if settings.MSSQL else psql_code,
                          reverse_sql='ALTER TABLE [tblClaimServices] ALTER COLUMN [JsonExt] TEXT'
                          if settings.MSSQL else psql_code),
        migrations.RunSQL('ALTER TABLE [tblClaimItems] ALTER COLUMN [JsonExt] NVARCHAR(MAX)'
                          if settings.MSSQL else psql_code,
                          reverse_sql='ALTER TABLE [tblClaimItems] ALTER COLUMN [JsonExt] TEXT'
                          if settings.MSSQL else psql_code),
    ]



# Generated by Django 2.1.11 on 2019-08-21 06:34

import core.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BatchRun',
            fields=[
                ('id', models.AutoField(db_column='RunID', primary_key=True, serialize=False)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('run_date', core.fields.DateTimeField(db_column='RunDate')),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(db_column='AuditUserID')),
                ('run_year', models.IntegerField(db_column='RunYear')),
                ('run_month', models.SmallIntegerField(db_column='RunMonth')),
            ],
            options={
                'db_table': 'tblBatchRun',
                'managed': False,
            },
        ),
    ]



# Generated by Django 3.0.14 on 2022-08-22 18:48

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0027_auto_20240203_2339'),
    ]

    operations = [
        migrations.CreateModel(
            name='ClaimServiceService',
            fields=[
                ('idCss', models.AutoField(db_column='idCss', primary_key=True, serialize=False)),
                ('qty_provided', models.IntegerField(blank=True, db_column='qty_provided', null=True)),
                ('qty_displayed', models.IntegerField(blank=True, db_column='qty_displayed', null=True)),
                ('pcpDate', models.DateTimeField(blank=True, db_column='created_date', default=django.utils.timezone.now, null=True)),
                ('price_asked', models.DecimalField(blank=True, db_column='price', decimal_places=2, max_digits=18, null=True)),
                ('claimlinkedService', models.ForeignKey(db_column='claimlinkedService', on_delete=django.db.models.deletion.DO_NOTHING, related_name='claimlinkedService', to='claim.ClaimService')),
                ('service', models.ForeignKey(db_column='ServiceId', on_delete=django.db.models.deletion.DO_NOTHING, related_name='claimServices', to='medical.Service')),
            ],
            options={
                'db_table': 'tblClaimServicesService',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='ClaimServiceItem',
            fields=[
                ('idCsi', models.AutoField(db_column='idCsi', primary_key=True, serialize=False)),
                ('qty_provided', models.IntegerField(blank=True, db_column='qty_provided', null=True)),
                ('qty_displayed', models.IntegerField(blank=True, db_column='qty_displayed', null=True)),
                ('pcpDate', models.DateTimeField(blank=True, db_column='created_date', default=django.utils.timezone.now, null=True)),
                ('price_asked', models.DecimalField(blank=True, db_column='price', decimal_places=2, max_digits=18, null=True)),
                ('claimlinkedItem', models.ForeignKey(db_column='ClaimServiceID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='claimlinkedItem', to='claim.ClaimService')),
                ('item', models.ForeignKey(db_column='ItemID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='claimItems', to='medical.Item')),
            ],
            options={
                'db_table': 'tblClaimServicesItems',
                'managed': True,
            },
        ),
    ]



# Generated by Django 3.0.3 on 2020-11-26 12:44

import core.fields
import core.utils
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0010_auto_20200611_0601'),
    ]

    operations = [
        migrations.AlterField(
            model_name='claimattachment',
            name='date',
            field=core.fields.DateField(blank=True, default=core.utils.TimeUtils.now),
        ),
    ]
